{
  Word mail merge component.
  Made by Robert van Tilburg, August 2000
  lothar@lucrasoft.nl
  www.robertvantilburg.nl
}
unit MailMerge;

interface

uses
  Classes, Word97, Sysutils;

type
  EMailMergeError = class(Exception)
  end;

  TMailMerge = Class(TComponent)
  private
    FWordApp: TWordApplication;
    FWordDoc: _Document;
    FColumns: TStringList;
    FGrid: array of array of String;
    FRecordNr: Integer;
    FDocumentName: TFileName;
    FStatusMessage: String;
    FOnStatusMessage: TNotifyEvent;
    function    GetColumns: String;
    procedure   SetColumns(const Value: String);
    function    GetFieldValue(const FieldName: string): String;
    procedure   SetFieldValue(const FieldName, Value: String);
    function    GetColumnNrByName(const ColumnName: String): Integer;
    procedure   SetDocumentName(const Value: TFileName);
    procedure   SetStatusMessage(const Value: String);
    function GetWordDoc: _Document;
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Append;
    procedure   Clear;
    procedure   Merge;
    property    FieldValues[const FieldName: string]: String read GetFieldValue write SetFieldValue; default;
    property    DocumentName: TFileName read FDocumentName write SetDocumentName;
    property    StatusMessage: String read FStatusMessage;
    property    WordApplication: TWordApplication read FWordApp;
    property    WordDocument: _Document read GetWordDoc;
  published
    property    Columns: String read GetColumns write SetColumns;
    property    OnStatusMessage: TNotifyEvent read FOnStatusMessage write FOnStatusMessage;
  end;

implementation

{ TMailMerge }

constructor TMailMerge.Create(AOwner: TComponent);
begin
  inherited;
  FWordApp := TWordApplication.Create(Self);
  FColumns := TStringList.Create;
  Clear;
end;

destructor TMailMerge.Destroy;
begin
  Clear;
  FColumns.Free;
  inherited;
end;

function TMailMerge.GetColumns: String;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to FColumns.Count - 1 do
  begin
    if Result = '' then
      Result := Result + FColumns[I]
    else
      Result := Result + FColumns[I] + ';';
  end;
end;

procedure TMailMerge.SetColumns(const Value: String);
var
  Start, I: Integer;
begin
  Start := 1;
  for I := 1 to Length(Value) do
  begin
    if Value[I] = ';' then
    begin
      FColumns.Add(Trim(Copy(Value, Start, (I - Start))));
      Start := I + 1;
    end;
  end;
  if Value[Length(Value)] <> ';' then
    FColumns.Add(Trim(Copy(Value, Start, MaxInt)));
end;

function TMailMerge.GetFieldValue(const FieldName: String): String;
var
  I: Integer;
begin
  if Length(FGrid) = 0 then
    raise EMailMergeError.Create('No current record');
  I := GetColumnNrByName(FieldName);
  Result := FGrid[FRecordNr, I];
end;

procedure TMailMerge.SetFieldValue(const FieldName, Value: String);
var
  I: Integer;
begin
  if Length(FGrid) = 0 then
    raise EMailMergeError.Create('No current record');
  I := GetColumnNrByName(FieldName);
  FGrid[FRecordNr, I] := Value;
end;

procedure TMailMerge.Clear;
var
  I: Integer;
begin
  for I := 0 to High(FGrid) do
    SetLength(FGrid[I], 0);
  SetLength(FGrid, 0);
  FColumns.Clear;
  FRecordNr := 0;
  FDocumentName := '';
  FStatusMessage := 'Idle';
end;

function TMailMerge.GetColumnNrByName(const ColumnName: String): Integer;
begin
  for Result := 0 to FColumns.Count - 1 do
    if (AnsiCompareText(FColumns[Result], ColumnName) = 0) then
      Exit;
  Raise EMailMergeError.CreateFmt('Column name does not exist: %s', [ColumnName]);
end;

procedure TMailMerge.Append;
begin
  SetLength(FGrid, Length(FGrid) + 1);
  FRecordNr := High(FGrid);
  SetLength(FGrid[FRecordNr], FColumns.Count);
end;

procedure TMailMerge.SetDocumentName(const Value: TFileName);
begin
  if (Length(Value) >= 2) and ((Value[2] <> ':') and (Value[1] + Value[2] <> '\\')) then
    Raise EMailMergeError.CreateFmt('document file "%s" must be a full path and file name', [Value]);
  if not FileExists(Value) then
    Raise EMailMergeError.CreateFmt('Document file not found: %s', [Value]);
  FDocumentName := Value;
end;

procedure TMailMerge.Merge;
var
  DataSourceFile, FileFormat, AddToRecentFiles: OleVariant;
  FileName, Pause: OleVariant;
  Doc: _Document;
  X, Y: Integer;
  T: Table;
begin
  FWordDoc := nil;
  if FDocumentName = '' then
    Raise EMailMergeError.Create('Document filename not set.');
  DataSourceFile := IncludeTrailingBackslash(ExtractFilePath(FDocumentName)) + 'MailMergeDatasource.temporary.doc';
  if (FileExists(DataSourceFile)) and (not DeleteFile(DataSourceFile)) then
    Raise EMailMergeError.CreateFmt('Unable to delete file %s', [DataSourceFile]);

  // Create the data source:
  SetStatusMessage('Creating datasource');

  Doc := FWordapp.Documents.Add(EmptyParam, EmptyParam);

  Doc.Tables.Add(Doc.Application.Selection.Range, Length(FGrid) + 1, FColumns.Count);
  T := Doc.Tables.Item(1);

  T.Rows.Item(1).Cells.Shading.BackgroundPatternColorIndex := wdGray25;
  T.Rows.Item(1).Range.Bold := Integer(True);


  // Fill columns on 1st row with column names:
  SetStatusMessage('Filling datasource with fields');
  for X := 0 to FColumns.Count - 1 do
  begin
    T.Columns.Item(X + 1).SetWidth(100, wdAdjustNone);

    T.Rows.Item(1).Cells.Item(X + 1).Range.Text := FColumns[X];
  end;

  // Fill cells of table:
  for Y := 0 to High(FGrid) do
  begin
    for X := 0 to FColumns.Count - 1 do
    begin
      T.Rows.Item(Y + 2).Cells.Item(X + 1).Range.Text := FGrid[Y][X];
    end;
  end;

  // Save the data source:
  SetStatusMessage('Saving datasource');
  FileFormat := wdFormatDocument;
  AddToRecentFiles := False;
  Doc.SaveAs(DataSourceFile, // filename
             FileFormat, // fileformat
             EmptyParam, // Lockcomments
             EmptyParam, // Password
             AddtoRecentFiles, // AddtoRecentFiles
             EmptyParam, // WritePassword
             EmptyParam, // ReadOnlyRecommended
             EmptyParam, // EmbedTrueTypeFonts
             EmptyParam, // SaveNativePictureFormat
             EmptyParam, // SaveFormsData
             EmptyParam  // SaveAsAOCELetter
            );
  SetStatusMessage('Closing datasource');
  Doc.Close(EmptyParam, EmptyParam, EmptyParam);

  // Load template:
  SetStatusMessage('Opening template');
  FileName := FDocumentName;
  Doc := FWordApp.Documents.Open(FileName,   // FileName
                                 EmptyParam, // ConfirmConversion
                                 EmptyParam, // ReadOnly
                                 EmptyParam, // AddToRecentFiles
                                 EmptyParam, // PasswordDocument
                                 EmptyParam, // PasswordTemplate
                                 EmptyParam, // Revert
                                 EmptyParam, // WritePasswordDocument
                                 EmptyParam, // WritePasswordTemplate
                                 EmptyParam  // Format
                                );

  SetStatusMessage('Showing Word');
  FWordApp.Visible := True;

  SetStatusMessage('Opening datasource');
  Doc.MailMerge.OpenDataSource(DataSourceFile,
                               FileFormat,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam,
                               EmptyParam);

  Doc.MailMerge.Destination := wdSendToNewDocument;
  Pause := False;

  SetStatusMessage('Merging');
  Doc.MailMerge.Execute(Pause);
  FWordDoc := FWordApp.ActiveDocument;

  SetStatusMessage('Closing template');
  Doc.Close(EmptyParam, EmptyParam, EmptyParam);
  DeleteFile(DataSourceFile);

  FWordDoc.Activate;

  SetStatusMessage('Done!');
end;

procedure TMailMerge.SetStatusMessage(const Value: String);
begin
  FStatusMessage := Value;
  if Assigned(FOnStatusMessage) then
    FOnStatusMessage(Self);
end;

function TMailMerge.GetWordDoc: _Document;
begin
  Result := FWordDoc;
  if Result = nil then
    Raise EMailMergeError.Create('You can only access the WordDoc interface after a successful call to the Merge method.');
end;

end.
